from __future__ import division

import sys
import datetime
import time
import re
import itertools
from collections import defaultdict


import parsedatetime


### pypy detection

try:
    import __pypy__
except ImportError:
    is_pypy = False
else:
    is_pypy = True



### py2-3 compatability


is_py3 = sys.version_info[0] == 3


if is_py3:
    bin_type = bytes
    text_type = str
else:
    bin_type = str
    text_type = unicode


def is_basestring(s):
    return isinstance(s, (bin_type, text_type))


if is_py3:
    range = range
else:
    range = xrange


if is_py3:
    itervalues = lambda d: d.values()
    iteritems = lambda d: d.items()
    iterkeys = lambda d: d.keys()

    def values(a_dict):
        return list(a_dict.values())

    def items(a_dict):
        return list(a_dict.items())

    def keys(a_dict):
        return list(a_dict)

else:
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()
    iterkeys = lambda d: d.iterkeys()

    def values(a_dict):
        return a_dict.values()

    def items(a_dict):
        return a_dict.items()

    def keys(a_dict):
        return list(a_dict)




if is_py3:

    class CmpMixin(object):

        def __eq__(self, other):
            return self.__cmp__(other) == 0

        def __ne__(self, other):
            return self.__cmp__(other) != 0

        def __gt__(self, other):
            return self.__cmp__(other) > 0

        def __lt__(self, other):
            return self.__cmp__(other) < 0

        def __ge__(self, other):
            return self.__cmp__(other) >= 0

        def __le__(self, other):
            return self.__cmp__(other) <= 0


    def cmp(a, b):

        if a is None and b is None:
            return 0
        if a is None and b is not None:
            return -1
        if a is not None and b is None:
            return 1

        if a == b:
            return 0

        # Python 2 allows comparing dictionaries, but it looks to be accidental
        if isinstance(a, dict) and isinstance(b, dict):
            return cmp(id(a), id(b))

        if a > b:
            return 1
        if a < b:
            return -1
        raise AssertionError('Invalid comparison')

else:

    class CmpMixin(object):
        pass

    cmp = cmp


NoneType = type(None)


### misc


undefined = object()


class cached_property(property):

    def __init__(self, func):
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.func = func

    def __set__(self, obj, value):
        obj.__dict__[self.__name__] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, undefined)
        if value is undefined:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value

    def __delete__(self, obj):
        if obj is None:
            return
        value = obj.__dict__.get(self.__name__, undefined)
        if value is not undefined:
            del obj.__dict__[self.__name__]



### strings and numbers

def get_num(x, num_fun, default):
    if x is None:
        return default
    try:
        return num_fun(x)
    except ValueError:
        return default

def get_float(x, default=None):
    return get_num(x, float, default)

def get_int(x, default=None):
    return get_num(x, int, default)

def partition_numstr(s):
    for i, c in enumerate(s):
        if not c.isdigit() and c not in {'.', ','}:
            k = i
            break
    else:
        k = len(s)
    return (get_float(s[:k].replace(',', '.')), s[k:])

def is_number(x):
    return isinstance(x, (int, float)) and not isinstance(x, bool)

_re_find_number = re.compile(r'[-+]?\d+[\.]?\d*')
def find_number(s):
    m = first(_re_find_number.finditer(s))
    if m is None:
        return None
    matched_s = m.group()
    if '.' in matched_s:
        return float(matched_s)
    return int(matched_s)

def find_all_number_strs(s):
    return _re_find_number.findall(s)

def avg(seq):
    s = 0
    l = 0
    for x in seq:
        s += x
        l += 1
    return s / l

def expand_tabs(astring, tablen=8):
    result = []
    for piece in astring.split('\t'):
        result.append(piece)
        result.append(' '*(tablen-len(piece)%tablen))
    return ''.join(result[:-1])

def strip(seq, to_strip, from_start=True, from_end=True):
    start = 0
    if from_start:
        while True:
            if start == len(seq):
                return []
            if not to_strip(seq[start]):
                break
            start += 1
    end = len(seq)-1
    if from_end:
        while True:
            if not to_strip(seq[end]):
                break
            end -= 1
    return seq[start:end+1]

def change_at(s, i, newsubstring):
    return s[:i] + newsubstring + s[i+1:]

def has_content(s):
    return s and any(c is not None and c.isalnum() for c in s)

def chars_positions(chars, s):
    chars_set = set(chars)
    res = defaultdict(list)
    for i in range(len(s)):
        if s[i] in chars_set:
            res[s[i]].append(i)
    return res

def chars_ranges(chars, s):
    cp = chars_positions(chars, s)
    return {c: list(ranges(cp[c])) for c in cp}



### lists and dicts

def first(iterable, default=None, key=None):
    if key is None:
        for el in iterable:
            if el:
                return el
    else:
        for el in iterable:
            if key(el):
                return el
    return default

def nestedget(o, *keys):
    for k in keys:
        try:
            o = o[k]
        except (IndexError, KeyError, TypeError):
            return None
    return o

def translate_indexes_after_deletion(idxs, deleted_idxs):
    if not idxs:
        return []
    max_idx = max(idxs)
    deleted_idxs = [idx for idx in deleted_idxs if idx <= max_idx]
    res = []
    for idx in idxs:
        if idx in deleted_idxs:
            continue
        count_before = count(iidx for iidx in deleted_idxs if iidx < idx)
        res.append(idx - count_before)
    return res

def uniq_sameorder(lst, key=lambda x:x):
    seen = set()
    return [seen.add(key(obj)) or obj for obj in lst if key(obj) not in seen]

def remove_indexes(lst, indexes):
    if not indexes:
        return lst
    indexes = set(indexes)
    return [x for i, x in enumerate(lst) if i not in indexes]

def fill_to_len(lst, wanted_len, fill_gen):
    for i in range(len(lst), wanted_len):
        lst.append(fill_gen())

def chunks(l, n, fill_gen=None):
    for i in range(0, len(l), n):
        chunk = l[i:i+n]
        if fill_gen is not None:
            fill_to_len(chunk, n, fill_gen)
        yield chunk

def count(seq):
    if isinstance(seq, (list, tuple)):
        return len(seq)
    res = 0
    for x in seq:
        res += 1
    return res

def ranges(i):
    for a, b in itertools.groupby(enumerate(i), lambda x__y: x__y[1] - x__y[0]):
        b = list(b)
        yield b[0][1], b[-1][1]+1

def flatten(seq):
    return [el for subseq in seq for el in subseq]

def dict_copy_with_item(d, k, v):
    res = d.copy()
    res[k] = v
    return res

def format_rows(rows):
    lines = []
    for row in rows:
        r = repr(row)
        lines.append(r)
    lines_repr = ',\n'.join(lines)
    return '[\n%s\n]\n' % lines_repr



### datetime

_pdt_cal = parsedatetime.Calendar()
def parsedatetime_parse(s):
    st = _pdt_cal.parse(s)[0]
    return datetime_from_struct_time(st)

def datetime_from_struct_time(st):
    return datetime.datetime.fromtimestamp(time.mktime(st))

